/*
 * pgm.h - 000330
 * Copyright (c) 2000 Gianluca Iannaccone, Luigi Rizzo
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR `AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#ifndef ns_pgm_h
#define ns_pgm_h

#include "agent.h"
#include "packet.h"
#include "rng.h"
#include "timer-handler.h"


// Base PGM Agent

class PGMAgent : public Agent {
public:
  PGMAgent();
  virtual ~PGMAgent() { }
  virtual int command(int argc, const char*const* argv);
  virtual void recv(Packet* p, Handler*) { Packet::free(p); } ;

  inline int pgm_seq_lt(u_int a, u_int b) { return  ((int) (a - b) < 0) ; } 
  inline int pgm_seq_leq(u_int a, u_int b) { return  ((int) (a - b) <= 0) ; } 
  inline int pgm_seq_gt(u_int a, u_int b) { return  ((int) (a - b) > 0) ; } 
  inline int pgm_seq_geq(u_int a, u_int b) { return  ((int) (a - b) >= 0) ; } 

protected:
  RNG rng_; // random number generator

  // packet header offsets
  int off_pgm_;
  int off_pgm_spm_;
  int off_pgm_odata_;
  int off_pgm_rdata_;
  int off_pgm_nak_;
  int off_pgm_nnak_;
  int off_pgm_ncf_;
  // pgmcc 
  int off_pgm_ack_ ;

  // utility functions to print statistics
  virtual void print_statistics(void) ;
  virtual void print_xplot(Packet *) ;

  // statistics
  Tcl_Channel xplotchannel_ ;
  int color_ ;  // color for ODATA pkts in xplot graphs
  unsigned int pkts_sent_;
  unsigned int pkts_rcvd_;
  unsigned int odata_sent_;
  unsigned int odata_rcvd_;
  unsigned int rdata_sent_;
  unsigned int rdata_rcvd_;
  unsigned int spm_sent_;
  unsigned int spm_rcvd_;
  unsigned int nak_sent_;
  unsigned int nak_rcvd_;
  unsigned int ncf_sent_;
  unsigned int ncf_rcvd_;
  unsigned int nnak_sent_;
  unsigned int nnak_rcvd_;
  unsigned int ack_sent_;
  unsigned int ack_rcvd_;
  unsigned int unrecoverable_pkts_;
};

// Packet Headers 

// PGM Common header

struct hdr_pgm {
  nsaddr_t tsi_;
  unsigned int sqn_;
  unsigned char options_;
#define PGM_OPT_PRESENT         0x01    // there are option extensions
#define PGM_OPT_NE_PRESENT      0x02    // there are NE-sign. opt. ext.
#define PGM_OPT_PARITY          0x80    // this is a parity pkt

  // options
  unsigned int opttype_ ;
  // OPT_JOIN
  unsigned int opt_join_min_ ;          // min seqno for repair

  static int offset_;

  // per field member functions
  nsaddr_t& tsi() {return tsi_;}
  unsigned int& sqn() {return sqn_;}
  unsigned char& options() {return options_;}
  unsigned int& opttype() {return opttype_;}
  unsigned int& opt_join_min() {return opt_join_min_;}
  inline static int& offset() { return offset_; }
  inline static hdr_pgm* access(Packet* p) {
        return (hdr_pgm*) p->access(offset_); }
};

// SPM header

#define HDR_SIZE_PGM_SPM	36	// simulated header size
struct hdr_pgm_spm {
  unsigned int trail_;
  unsigned int lead_;
  nsaddr_t path_nla_;
  static int offset_;

  // per field member functions
  unsigned int& trail() {return trail_;}
  unsigned int& lead() {return lead_;}
  nsaddr_t& path_nla() {return path_nla_;}
  inline static int& offset() { return offset_; }
  inline static hdr_pgm_spm* access(Packet* p) {
        return (hdr_pgm_spm*) p->access(offset_); }
};

// ODATA header 

#define HDR_SIZE_PGM_ODATA	24	// simulated header size
struct hdr_pgm_odata {
  unsigned int trail_;

  // options 
  nsaddr_t acker_ ; // acker address (for pgmcc)
  double timestamp_ ; // packet timestamp (for pgmcc) 

  static int offset_;

  // per field member functions
  unsigned int& trail() {return trail_;}
  nsaddr_t& acker() {return acker_;}
  double& timestamp() {return timestamp_;} 
  inline static int& offset() { return offset_; }
  inline static hdr_pgm_odata* access(Packet* p) {
        return (hdr_pgm_odata* ) p->access(offset_); }
};

// RDATA header

#define HDR_SIZE_PGM_RDATA	24	// simulated header size
struct hdr_pgm_rdata {
  unsigned int trail_;
  static int offset_;

  // per field member functions
  unsigned int& trail() {return trail_;}
  inline static int& offset() { return offset_; }
  inline static hdr_pgm_rdata* access(Packet* p) {
        return (hdr_pgm_rdata* ) p->access(offset_); }
};

// NAK header

#define HDR_SIZE_PGM_NAK	36	// simulated header size
struct hdr_pgm_nak {
  nsaddr_t src_nla_;
  static int offset_;

  // options
  double rx_loss_ ;	// loss rate observed by recv (for pgmcc) 
  u_int rx_lead_ ;	// rxw_lead_ (for pgmcc) 
  nsaddr_t nacker_ ;    // host that generated the NAK (for pgmcc)
  double timestamp_ ;   // last received ODATA packet timestamp (for pgmcc) 

  // per field member functions
  nsaddr_t& src_nla() {return src_nla_;}
  double& rx_loss() {return rx_loss_;}
  u_int& rx_lead() {return rx_lead_;}
  nsaddr_t& nacker() {return nacker_;}
  double& timestamp() {return timestamp_;}
  inline static int& offset() { return offset_; }
  inline static hdr_pgm_nak* access(Packet* p) {
        return (hdr_pgm_nak*) p->access(offset_); }
};

// NNAK header 

#define HDR_SIZE_PGM_NNAK	36	// simulated header size
struct hdr_pgm_nnak {
  nsaddr_t src_nla_;
  static int offset_;
 
  // per field member functions
  nsaddr_t& src_nla() {return src_nla_;}
  inline static int& offset() { return offset_; }
  inline static hdr_pgm_nnak* access(Packet* p) {
        return (hdr_pgm_nnak*) p->access(offset_); }
};

// NCF header

#define HDR_SIZE_PGM_NCF	36	// simulated header size
struct hdr_pgm_ncf {
  nsaddr_t src_nla_;
  static int offset_;
 
  // per field member functions
  nsaddr_t& src_nla() {return src_nla_;}
  inline static int& offset() { return offset_; }
  inline static hdr_pgm_ncf* access(Packet* p) {
        return (hdr_pgm_ncf*) p->access(offset_); }
};
 
// ACK header (for pgmcc) 

#define HDR_SIZE_PGM_ACK	36	// simulated header size
struct hdr_pgm_ack {
  u_int ack_rxw_lead_ ;
  u_int32_t bitmask_ ;
  double rx_loss_ ;
  nsaddr_t nacker_ ;
  double timestamp_ ;		
 
  static int offset_;
 
  // per field member functions
  u_int32_t& bitmask() { return bitmask_ ; }
  double& rx_loss() { return rx_loss_ ; }
  u_int& ack_rxw_lead() { return ack_rxw_lead_ ; }
  nsaddr_t& nacker() { return nacker_ ; }
  double& timestamp() { return timestamp_ ; }
  inline static int& offset() { return offset_; }
  inline static hdr_pgm_ack* access(Packet* p) {
        return (hdr_pgm_ack*) p->access(offset_); }
};

// PGM Options

#define PGM_OPT_FRAGMENT        0x00000001      // unimplemented
#define PGM_OPT_JOIN            0x00000002
#define PGM_OPT_TIME            0x00000004      // unimplemented
#define PGM_OPT_RXQ             0x00000008      // unimplemented
#define PGM_OPT_DROP            0x00000010      // unimplemented
#define PGM_OPT_REDIRECT        0x00000020      // unimplemented
// experimental options
#define PGM_OPT_DESC            0x00000040      // unimplemented
#define PGM_OPT_SYN             0x00000080      // unimplemented
#define PGM_OPT_FIN             0x00000100      // unimplemented
// simulated option size
#define OPT_LENGTH_SIZE		4
#define OPT_JOIN_SIZE       	8
 
// pgmcc options
#define PGM_OPT_LOSSRATE        0x00010000		// loss report present
#define PGM_OPT_SEND_NAK        0x00020000    	// receivers, please send a NAK
#define PGM_OPT_ELECT           0x00040000    	// node X is elected as acker
#define PGM_OPT_TIMESTAMP       0x00080000    	// timestamp present
// simulated option size
#define OPT_LOSSRATE_SIZE       8
#define OPT_SEND_NAK_SIZE       4
#define OPT_ELECT_SIZE          4
#define OPT_TIMESTAMP_SIZE      4

// Timers

#define NEXT_PACKET 	0
#define AMBIENT_SPM 	1
#define TXW_ADVANCE 	2

#define NAK_TIMEOUT 	3
#define NCF_TIMEOUT 	4
#define RDATA_TIMEOUT 	5

#define CC_TIMEOUT	6

class PGMTimer : public TimerHandler {
public:
  PGMTimer(PGMAgent *a, int t) : TimerHandler(), agent_(a), timer_type_(t) {}

protected:
  virtual void expire(Event *) { agent_->timeout(timer_type_) ; }
  PGMAgent *agent_;
  int timer_type_;
};

#endif // ns_pgm_h
