#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "tclcl.h"
#include "agent.h"
#include "address.h"
#include "ip.h"
#include "trace.h"
#include "classifier.h"
#include "my_pgmnet.h"

static class DemuxReplicatorClass : public TclClass {
public:
  DemuxReplicatorClass() : TclClass("Classifier/Replicator/Demuxer") {}
  TclObject* create(int, const char*const*) {
    return (new DemuxReplicator());
  }
} class_demux_replicator;

static class PGMNetAgentClass : public TclClass {
public:
  PGMNetAgentClass() : TclClass("Classifier/Replicator/Demuxer/PGM") {}
  TclObject* create(int, const char*const*) {
    return ((DemuxReplicator*)new PGMNetAgent());
  }
} class_pgm_net_agent;


// PGM Net Element Rtx State Info

PGMNetRtxInfo::PGMNetRtxInfo(PGMAgent *a, nsaddr_t t, unsigned int s, nsaddr_t n, double loss, unsigned int lead) : 
  sqn_(s), tsi_(t), src_nla_(n), rx_loss_(loss), rx_lead_(lead),
  nak_repeat_timer_(a, PGM_NET_TIMER_NAK_REPEAT, s),
  nak_wait_for_ncf_timer_(a, PGM_NET_TIMER_WAIT_FOR_NCF, s),
  nak_elim_timer_(a, PGM_NET_TIMER_NAK_ELIM, s),
  keep_rtx_state_timer_(a, PGM_NET_TIMER_KEEP_RTX_STATE, s)
{
}

PGMNetRtxInfo::~PGMNetRtxInfo() {

  if (nak_repeat_timer_.status() == TIMER_PENDING)
    nak_repeat_timer_.cancel();
  if (nak_wait_for_ncf_timer_.status() == TIMER_PENDING)
    nak_wait_for_ncf_timer_.cancel();
  if (nak_elim_timer_.status() == TIMER_PENDING)
    nak_elim_timer_.cancel();
  if (keep_rtx_state_timer_.status() == TIMER_PENDING)
    keep_rtx_state_timer_.cancel();
}
// PGM Network Element Agent

PGMNetAgent::PGMNetAgent() :  spm_path_(0)
{
  Replicator::bind("nak_repeat_time_", &nak_repeat_time_);
  Replicator::bind("nak_wait_for_ncf_time_", &nak_wait_for_ncf_time_);
  Replicator::bind("nak_elim_time_", &nak_elim_time_);
  Replicator::bind("keep_rtx_state_time_", &keep_rtx_state_time_);
  Replicator::bind("pgm_group_", &pgm_group_);
  PGMAgent::bind("addr_",&here_.addr_);

  rtxstate_ = new Tcl_HashTable;
  Tcl_InitHashTable(rtxstate_, TCL_ONE_WORD_KEYS);

}

int PGMNetAgent::command(int argc, const char*const* argv)
{
  int res;
  res = Replicator::command(argc,argv);
  if (res==TCL_OK) 
	return res;
  else
  	return PGMAgent::command(argc, argv);
}

void PGMNetAgent::timeout(int tno, unsigned int sqn) {

  Packet *p;
  PGMNetRtxInfo *rtxinfo = rtxstate_lookup(sqn);
 
  switch(tno) {

  case PGM_NET_TIMER_NAK_REPEAT:

    p = allocnakpkt(rtxinfo->tsi(), rtxinfo->sqn(), rtxinfo->src_nla(), rtxinfo->rx_loss(), rtxinfo->rx_lead());
    rtxinfo->nak_repeat_timer().resched(nak_repeat_time_);
    send(p, 0);
    break;

  case PGM_NET_TIMER_WAIT_FOR_NCF:

    cout << "PGM Warning: Cancelling NAK forwarding due to missing NCF (sqn " <<
      rtxinfo->sqn() << ")" << endl;
    rtxstate_delete(sqn);
    break;

  case PGM_NET_TIMER_NAK_ELIM:
    // don't need to do anything - recv_nak() checks the status() of this timer
    break;

  case PGM_NET_TIMER_KEEP_RTX_STATE:

    cout << "PGM Warning: Cancelling NAK forwarding due to missing RDATA (sqn " <<
     rtxinfo->sqn() << ")" << endl;
    rtxstate_delete(sqn);
    break;

  default:

    cout << "PGM Warning: Unknown timer expired" << endl;
    break;

  }
}

Packet *PGMNetAgent::allocnakpkt(nsaddr_t nak_tsi, unsigned int nak_sqn, 
				 nsaddr_t src_nla, double loss, unsigned int lead) {
  // allocate new NAK packet
  size_ = HDR_SIZE_PGM_NAK;
  type_ = PT_PGM_NAK;
  Packet *p = allocpkt();

  // fill in fields of NAK packet
  hdr_ip *ih = hdr_ip::access(p) ;
  hdr_pgm *ph = (hdr_pgm *) p->access(off_pgm_);
  hdr_pgm_nak *pnh = (hdr_pgm_nak *) p->access(off_pgm_nak_);
  ih->dst().addr_ = spm_path_;
  ph->tsi() = nak_tsi;
  ph->sqn() = nak_sqn;
  ph->options() = 0;
  pnh->src_nla() = src_nla;
  pnh->rx_loss() = loss;
  pnh->rx_lead()= lead;

  return p;
}

PGMNetRtxInfo *PGMNetAgent::rtxstate_init(nsaddr_t t, unsigned int s, 
					  nsaddr_t n, double loss, unsigned int lead) {
  int new_entry = 0;
  Tcl_HashEntry *he = Tcl_CreateHashEntry(rtxstate_, (char *) s, &new_entry);

  if (!new_entry) {// we already have rtx state for this sqn
    cout << "PGM Error: Rtx info already exists for sqn " << s << endl;
    exit(-1);
  }

  PGMNetRtxInfo *rtxinfo = new PGMNetRtxInfo(this, t, s, n, loss, lead);
  Tcl_SetHashValue(he, (ClientData *) rtxinfo);
  return rtxinfo;
}

PGMNetRtxInfo *PGMNetAgent::rtxstate_lookup(unsigned int s) {
  Tcl_HashEntry *he = Tcl_FindHashEntry(rtxstate_, (char *) s);
  if (he == NULL)
    return NULL;
  else
    return (PGMNetRtxInfo *) Tcl_GetHashValue(he);
}

void PGMNetAgent::rtxstate_delete(unsigned int s) {
  Tcl_HashEntry *he = Tcl_FindHashEntry(rtxstate_, (char *) s);
  if (he != NULL) {
    PGMNetRtxInfo *rtxinfo = (PGMNetRtxInfo *) Tcl_GetHashValue(he);
    delete rtxinfo;
    Tcl_DeleteHashEntry(he);
  }
}

void PGMNetAgent::recv(Packet *p, Handler *h) {

  hdr_cmn *ch = hdr_cmn::access(p);
  hdr_ip *ih = hdr_ip::access(p);
  hdr_pgm *ph = (hdr_pgm *) p->access(off_pgm_);
  hdr_pgm_spm *psh = (hdr_pgm_spm *) p->access(off_pgm_spm_);
  hdr_pgm_nak *pnh = (hdr_pgm_nak *) p->access(off_pgm_nak_);
  hdr_pgm_ncf *pnch = (hdr_pgm_ncf *) p->access(off_pgm_ncf_);
 
  PGMNetRtxInfo *rtxinfo;
  int must_send;

  switch (ch->ptype()) {

  case PT_PGM_ODATA:
  
	Replicator::recv(p, h);
	break;

  case PT_PGM_RDATA:	

  	rtxinfo = rtxstate_lookup(ph->sqn());
  	if (rtxinfo == NULL) 
    		Packet::free(p);
  	else {
    	// constrain PGM replicator to this sqn's oif list
    	Tcl::instance().evalf("%s disable-all-ifaces", Replicator::name());
    	while (rtxinfo->agent_exists())
      	Tcl::instance().evalf("%s enable-agent %d", Replicator::name(), 
			    rtxinfo->pop_agent());
    	while (rtxinfo->iface_exists())
      	Tcl::instance().evalf("%s enable-oif %d", Replicator::name(), 
			    rtxinfo->pop_iface());
      
  	Replicator::recv(p, h);

    	// reset PGM replicator to forward to all interfaces in oif list
    	Tcl::instance().evalf("%s enable-active-ifaces", Replicator::name());
    	rtxstate_delete(ph->sqn());
        }
	break;

  case PT_PGM_SPM:

  	spm_path_ = psh->path_nla();
  	psh->path_nla() = here_.addr_;
  	Replicator::recv(p, h);
	break;

  case PT_PGM_NAK:

	{
  	// send NCF downstream
  	size_ = HDR_SIZE_PGM_NCF;
  	type_ = PT_PGM_NCF;
  	Packet *p2 = allocpkt();
 	hdr_ip *ih2 = hdr_ip::access(p2);
  	hdr_pgm *ph2 = (hdr_pgm *) p2->access(off_pgm_);
  	hdr_pgm_ncf *pnch2 = (hdr_pgm_ncf *) p2->access(off_pgm_ncf_);

  	ih2->src().addr_ = ph->tsi() ;
  	ih2->dst().addr_ = pgm_group_ ;
  	ph2->tsi() = ph->tsi() ;
  	ph2->sqn() = ph->sqn() ;
  	ph2->options() = 0 ;
	pnch2->src_nla() = pnh->src_nla() ;

  	// constrain PGM replicator to this packet's iif
  	Tcl::instance().evalf("%s disable-all-ifaces", Replicator::name());
  	if (ch->iface() == UNKN_IFACE.value()) // this NAK is from a local agent
    	  Tcl::instance().evalf("%s enable-agent %d", Replicator::name(), ih->src().addr_);
  	else
    	  Tcl::instance().evalf("%s enable-oif %d", Replicator::name(), ch->iface());
 	Replicator::recv(p2, h);
  	// reset PGM replicator to forward to all interfaces in oif list
  	Tcl::instance().evalf("%s enable-active-ifaces", Replicator::name());
  	// update retransmission state and timers
  	rtxinfo = rtxstate_lookup(ph->sqn());

 	must_send=0;
  	if (rtxinfo == NULL) {
    	  rtxinfo = rtxstate_init(tsi_, ph->sqn(), pnh->src_nla(), pnh->rx_loss(), pnh->rx_lead());
    	  rtxinfo->nak_repeat_timer().sched(nak_repeat_time_);
    	  rtxinfo->ncf_wait_timer().sched(nak_wait_for_ncf_time_);
   	  must_send=1;
  	}
    	// update agent / outgoing interface list
    	if (ch->iface() == UNKN_IFACE.value()) // local agent
      	  rtxinfo->push_agent(ih->src().addr_);
    	else
      	  rtxinfo->push_iface(ch->iface());
  
        if (must_send) {
   	  // send NAK upstream
    	  ih->src() = here_;
          pnh->src_nla() = here_.addr_;
    	  ih->dst().addr_ = spm_path_;
    	  send(p, h);
  	} else  {
      	  // eliminate NAK
      	  Packet::free(p);
  	}
	break;
	}

  case PT_PGM_NNAK:

	break;

  case PT_PGM_NCF:
  
  	// update rtx state and timers
  	rtxinfo = rtxstate_lookup(ph->sqn());
  	if (rtxinfo == NULL) {
    	  // unsolicited NCF: do NAK anticipation
    	  rtxinfo = rtxstate_init(tsi_, ph->sqn(), pnch->src_nla(),0,0);
    	  rtxinfo->nak_elim_timer().sched(nak_elim_time_);
    	  rtxinfo->keep_rtx_state_timer().sched(keep_rtx_state_time_);
  	}
  	else {
    	  if (rtxinfo->nak_repeat_timer().status() == TIMER_PENDING)
      	  rtxinfo->nak_repeat_timer().cancel();
    	  if (rtxinfo->ncf_wait_timer().status() == TIMER_PENDING)
      	  rtxinfo->ncf_wait_timer().cancel();
    	  rtxinfo->nak_elim_timer().resched(nak_elim_time_);
    	  rtxinfo->keep_rtx_state_timer().resched(keep_rtx_state_time_);
  	}

  	Packet::free(p);
	break;

  default:
	printf("PGM Error:Unknown packet type \n");
	break;
  }
}