/*
 * pgmrcv.cc - 000330
 * Copyright (c) 2000 Gianluca Iannaccone, Luigi Rizzo
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR `AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdio.h>

#include "tclcl.h"
#include "agent.h"
#include "queue.h"
#include "packet.h"
#include "ip.h"
#include "my_pgmrcv.h"

static class PGMRcvAgentClass : public TclClass {
public:
  PGMRcvAgentClass() : TclClass("Agent/PGM/Rcv") {}
  TclObject* create(int, const char*const*) { return (new PGMRcvAgent()); }
} class_pgm_rcv_agent;

inline void 
PGMRetxTimer::expire(Event *) 
{
  agent_->timeout(nak_, timer_type_) ;
}

ReassemblyQueue::ReassemblyQueue() 
{
  seg_next_ = seg_prev_ = this ;
  nak_next_ = nak_prev_ = this ;
}

ReassemblyQueue::ReassemblyQueue(u_int seq, int size) 
{
  seg_next_ = seg_prev_ = NULL ;
  nak_next_ = nak_prev_ = NULL ;
  seq_ = seq ; 
  type_ = T_ODATA ;
  u.pkt_size_ = size ;
}

ReassemblyQueue::ReassemblyQueue(u_int seq, double time, PGMRcvAgent * a) :
timeout_(a, this, NAK_TIMEOUT) 
{
  seg_next_ = seg_prev_ = NULL ;
  nak_next_ = nak_prev_ = NULL ;     
  seq_ = seq ; 
  type_ = T_NAK ; 
  u.nak.ncf_retry_ = 0 ;
  u.nak.rdata_retry_ = 0 ;
  timeout_.sched(time) ;
}

void 
ReassemblyQueue::hole() 
{
  unlink_nak() ;
  timeout_.force_cancel() ;
  type_ = T_HOLE ;
}

void
ReassemblyQueue::fill(int size) 
{
  unlink_nak() ;
  timeout_.force_cancel() ;
  type_ = T_ODATA ;
  u.pkt_size_ = size ;
}

// PGM Receiver Agent

PGMRcvAgent::PGMRcvAgent() : 
state_(NEW),
path_nla_(0),
src_nla_(0),
rxw_trail_(0), 
rxw_lead_(0),
rxw_lookahead_(PGM_DEFAULT_LOOKAHEAD),
rxw_irs_(0),
rxw_next_(0),
recvspace_(30),
pkts_stored_(0)
{
  bind("nak_bo_ivl_",&nak_bo_ivl_) ;
  bind("ncf_to_ivl_",&ncf_to_ivl_) ;
  bind("rdata_to_ivl_",&rdata_to_ivl_) ;
  bind("ncf_retries_",&ncf_retries_) ;
  bind("rdata_retries_",&rdata_retries_) ;
}

void
PGMRcvAgent::timeout(ReassemblyQueue * q, int type) 
{
    int toomuch = 0 ;

    switch (type) {
    case NAK_TIMEOUT:
       /* Can send NAK only if have path_nla from an SPM */
       if ( path_nla_ ) {  
          output(PT_PGM_NAK, q->seq());
          q->timeout_.resched(ncf_to_ivl_, NCF_TIMEOUT) ;
       } else     /* retry later... */
          q->timeout_.resched(rng_.uniform(nak_bo_ivl_), NAK_TIMEOUT) ;
       return ;
    case NCF_TIMEOUT:
       toomuch = q->ncf_retry(ncf_retries_) ;
       break ;
    case RDATA_TIMEOUT:
       toomuch = q->rdata_retry(rdata_retries_) ;
       break ;
    } 
    
    if ( toomuch ) {
       fprintf(stderr, "-- [%5.3f] PGMRcvr: unrecov. loss due to "
               "missing NCF or RDATA (%d)\n", Scheduler::instance().clock(), 
               q->seq() );
       q->hole(); /* irrecov. loss due to missing NCF */
       clean_reass() ;
    } else  /* schedule nak ... */
       q->timeout_.resched(rng_.uniform(nak_bo_ivl_), NAK_TIMEOUT) ;
}

void
PGMRcvAgent::clean_reass() 
{ 
    while ( !reass_q_.empty() ) {
        ReassemblyQueue * q = reass_q_.seg_next_ ;
    
        // INVARIANT: at each stage, q->seq() == rxw_next_. If not
        // there is a bad mistake in the code.
        if (q->seq() != rxw_next_) { /* check invariant */
            fprintf(stderr, "-- [%5.3f] PGMRcvr: reass - "
                    "seq %u, rxw_next_ %d, rxw_trail_ %d, rxw_lead_ %d\n",
                    Scheduler::instance().clock(), q->seq(), 
                    rxw_next_, rxw_trail_, rxw_lead_);
            exit(-1) ;
        }

        if ( pgm_seq_lt(rxw_next_, rxw_trail_) && q->type() == T_NAK ) 
            q->hole() ; /* NAK expired for trail advance */

        if ( q->type() == T_NAK ) /* recoverable NAK */
            break ;
        else if ( q->type() == T_ODATA ) 
            pkts_stored_-- ;
        else if ( q->type() == T_HOLE )
            unrecoverable_pkts_++ ;
        
        // record gone. advance next,trail,unlink from queue
        rxw_next_++;
        if ( pgm_seq_lt(rxw_trail_, rxw_next_ ) )
            rxw_trail_ = rxw_next_ ;
        q->unlink() ;
        delete q ;
    }
}


void 
PGMRcvAgent::recv(Packet* p, Handler*)
{
  hdr_cmn *ch = hdr_cmn::access(p);
  hdr_pgm *ph = (hdr_pgm *) p->access(off_pgm_) ;
  u_int dp_seq ;

  if (ph->options() & PGM_OPT_PRESENT) 
     option_handler(p) ;

  if (state_ == NEW) {
     /* 
      * First SPM or ODATA makes receiver socket connected.
      */
     if ( ch->ptype() == PT_PGM_ODATA ) 
       dp_seq = ph->sqn() ; // packet sequence number 
     else if ( ch->ptype() == PT_PGM_SPM ) {
       hdr_pgm_spm *psh = (hdr_pgm_spm *) p->access(off_pgm_spm_) ;
       dp_seq = psh->lead() + 1 ; // XXX
     } else 
       goto fail ;

     state_ = CONNECTED ;
     src_nla_ = ph->tsi() ;  /* XXX next pkts should have same tsi */
     if ( rxw_irs_ == 0 ) 
        rxw_irs_ = rxw_trail_ = rxw_lead_ = rxw_next_ = dp_seq ;
  }

  u_int dp_trail ;
  u_int dp_lead ;
  hdr_pgm_odata * pdh ;
  hdr_pgm_spm * psh ;
  ReassemblyQueue * q ;
  ReassemblyQueue * r ; 
  u_int seq ;                // first packet we can recover

  switch(ch->ptype()) {
  case PT_PGM_NAK:
    fprintf(stderr, "PGMRcvAgent::recv, NAK received\n") ;
  case PT_PGM_NCF:
    dp_seq = ph->sqn() ;
    /* handle this for nak suppression */
    if ( pgm_seq_geq(dp_seq, rxw_trail_) && pgm_seq_leq(dp_seq, rxw_lead_) ) {
      ReassemblyQueue * q ;

      for ( q = reass_q_.nak_next_ ; q != &reass_q_ ; q = q->nak_next_) 
         if ( q->seq() == dp_seq ) {
            if ( ch->ptype() == PT_PGM_NCF ) {
               ncf_rcvd_++ ;
               q->timeout_.resched(rdata_to_ivl_, RDATA_TIMEOUT) ;
               q->reset_ncf_retries() ;
            } else {
               /* must be a NAK, so act as if the NAK had been sent */
               nak_rcvd_++ ;
               if ( q->timeout_.type() == NAK_TIMEOUT ) 
                  q->timeout_.resched(ncf_to_ivl_, NCF_TIMEOUT) ;
            }
            break; 
         }
    }
    goto done ;

  case PT_PGM_RDATA: 
  case PT_PGM_ODATA:
    pdh = (hdr_pgm_odata *) p->access(off_pgm_odata_) ;
    dp_trail = pdh->trail() ;
    dp_lead = dp_seq = ph->sqn() ;
   
#if 0
    // window size limitations in packets. XXX should be in bytes 
    if ( pkts_stored_ == recvspace_ ) 
       goto fail; /* receive window full */
#endif

    /* set lookahead to 1/2 of the current window or min 10 pkts */
    if ( pgm_seq_gt(rxw_lead_, rxw_trail_ + 2*PGM_DEFAULT_LOOKAHEAD) ) 
       rxw_lookahead_ = (rxw_lead_ - rxw_trail_)/2 ;
    else 
       rxw_lookahead_ = PGM_DEFAULT_LOOKAHEAD ; 
    
    if ( pgm_seq_lt(dp_seq, rxw_trail_) )  
       goto fail;  /* way too old */
 
    if ( pgm_seq_gt(dp_seq, rxw_lead_ + rxw_lookahead_) ) 
       /* new one.. might be rogue, check trail is in-window */
       if ( pgm_seq_lt(dp_trail, rxw_trail_) ||
            pgm_seq_gt(dp_trail, rxw_lead_ + rxw_lookahead_) ) 
          goto fail;

    if ( pgm_seq_gt(dp_seq, rxw_lead_) ) 
       rxw_lead_ = dp_seq ;
    if ( pgm_seq_gt(dp_trail, rxw_trail_) )
       rxw_trail_ = dp_trail ;
    if ( ch->ptype() == PT_PGM_ODATA ) 
       odata_rcvd_++ ;
    else 
       rdata_rcvd_++ ;
    break ;

  case PT_PGM_SPM: 
    psh = (hdr_pgm_spm *) p->access(off_pgm_spm_) ;
    dp_trail = psh->trail() ;
    dp_lead = psh->lead() ;

    spm_rcvd_++ ;
    path_nla_ = psh->path_nla() ;
    
    if ( pgm_seq_gt(dp_trail, rxw_trail_) ) {
       rxw_trail_ = dp_trail ;
       clean_reass() ;
    } 
  
    if ( pgm_seq_leq(dp_lead, rxw_lead_) ) 
       goto done ;

    /*
     * if i get here: dp_lead > rxw_lead_, and must insert entries
     * for NAK after rxw_lead_ .
     */
    dp_seq = rxw_lead_ = dp_lead ;
    break ;

  default:
    fprintf(stderr, "-- [%5.3f] PGMRcvr: discarding type %d\n",
            Scheduler::instance().clock(), ch->ptype()) ;
    goto fail;
  }

 /*
  * XXX at the moment we only get here with ODATA/RDATA/SPM. Should we
  * decide to use NAK/NCF to detect holes, make sure m is NULL so we
  * can tell the two cases.
  *
  * Locate place to insert. dp_seq is the current packet, q points
  * initially to the last record. After the scan, seq is the seqno of
  * the first missing packet, and q points to the record after which
  * we must insert new entries (which are seq..dp_seq inclusive).
  */ 
  
  q = reass_q_.seg_prev_ ;
  r = reass_q_.nak_prev_ ; 
  if ( reass_q_.empty() ) {
     seq = rxw_trail_ ;
     if ( seq > rxw_next_ ) 
        seq = rxw_next_ ;
  } else if ( pgm_seq_gt(dp_seq, q->seq()) ) {
     /* Pkt newer than last in queue, so start after that one. */
     seq = q->seq() + 1 ;
  } else if (ch->ptype() == PT_PGM_SPM) {  // XXX
     /* we are in the middle, but this is not data. */
     return ; /*XXX maybe should not happen ? */
  } else {
     /*
      * We are in the middle, need a full scan. Only follow the
      * NAK list, so skip over data pkts.
      */
     for ( r = reass_q_.nak_next_ ; r != &reass_q_ ; r = r->nak_next_ ) {
         /* XXX should exit when pgm_seq_gt( r->seq(), dp_seq); */
         if (r->seq() != dp_seq)
            continue;
         /* fill the hole */
         r->fill(ch->size()) ;
         goto present;
     }
     goto fail ;
  }

/*
  * insert an entry for each missing packet
  */
  if (ch->ptype() == PT_PGM_RDATA)
    goto fail;      /* should not get here!!! */
 
  for (; pgm_seq_leq(seq, dp_seq) ; seq++ ) {
    ReassemblyQueue * me ;
    if ( seq == dp_seq && ch->ptype() == PT_PGM_ODATA ) { 
        me = new ReassemblyQueue(seq, ch->size()) ;
        pkts_stored_++ ;
    } else {
        /* Insert nak entry */
        me = new ReassemblyQueue(seq, rng_.uniform(nak_bo_ivl_), this) ;
        me->link_nak(r) ;
        r = me ;
    } 

    /* insert into queue */
    me->link(q) ;
    q = me ;
  }


present:
  Packet::free(p) ; // i don't need it anymore
  clean_reass() ;
  return ; 

done:
fail:
  Packet::free(p) ; 
}

void
PGMRcvAgent::output(int type, u_int seq) 
{
  Packet * p ;

  switch (type) {
  case PT_PGM_NAK: 
    // send only if we know where is the next hop
   if ( path_nla_ ) {
       size_ = HDR_SIZE_PGM_NAK ;
       type_ = PT_PGM_NAK ;
       p = allocpkt() ;
       hdr_ip *ih = hdr_ip::access(p);
       hdr_pgm * ph = (hdr_pgm *) p->access(off_pgm_) ;
       hdr_pgm_nak * pnh = (hdr_pgm_nak *) p->access(off_pgm_nak_) ;

       ih->dst().addr_ = path_nla_ ;
       ph->tsi() = src_nla_ ;
       ph->sqn() = seq ;
       ph->options() = 0 ;
       pnh->src_nla() = here_.addr_ ;
       nak_sent_++ ;
       break ;
    }
    return ;
  default:
    break ;
  }

  pkts_sent_++ ;
  send(p,0) ;
}

void
PGMRcvAgent::buildoptions(Packet * p, packet_t) {
  hdr_cmn *ch = hdr_cmn::access(p);
  hdr_pgm * ph = (hdr_pgm *) p->access(off_pgm_) ;

  // no options implemented 

	// if options are present, put total length field
  if ( ph->options() & PGM_OPT_PRESENT ) 
     ch->size() += OPT_LENGTH_SIZE ;
}

void
PGMRcvAgent::option_handler(Packet * p) 
{
  hdr_pgm * ph = (hdr_pgm *) p->access(off_pgm_) ;

  if ( ph->opttype() & PGM_OPT_JOIN ) {
     ph->opttype() &= ~PGM_OPT_JOIN ;
     if ( state_ == NEW ) 
        rxw_trail_ = rxw_lead_ = rxw_next_ = rxw_irs_ = ph->opt_join_min() ; 
  } 

  if ( ph->opttype() ) 
     fprintf(stderr, "PGMRcvAgent::option_handler, unknown opt. (%x)\n", 
             ph->opttype()) ;
}