/*
 * pgmsnd.cc - 000330
 * Copyright (c) 2000 Gianluca Iannaccone, Luigi Rizzo
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR `AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdio.h>

#include "tclcl.h"
#include "agent.h"
#include "ip.h"
#include "my_pgmsnd.h"

static class PGMSndAgentClass : public TclClass {
public:
  PGMSndAgentClass() : TclClass("Agent/PGM/Snd") {}
  TclObject* create(int, const char*const*) {
    return (new PGMSndAgent());
  }
} class_pgm_snd_agent;

PGMSndAgent::PGMSndAgent() :  
  trail_advance_policy_(TRAIL_ADVANCE_TIMER), 
  app_waiting_(FALSE),
  pkt_waiting_(NULL),
  infinite_send_(FALSE),
  numbytes_(0),
  next_packet_timer_(this, NEXT_PACKET),
  txw_advance_timer_(this, TXW_ADVANCE),
  spm_ambient_timer_(this, AMBIENT_SPM)
{
  bind("txw_max_rte_", &txw_max_rte_) ;
  bind("txw_secs_", &txw_secs_) ;
  bind("spm_ambient_time_", &spm_ambient_time_) ;
  bind("sendspace_", &sendspace_) ;
  bind("payload_", &payload_) ;
  bind_bool("optjoin_", &optjoin_) ;

  txw_lead_ = 0 ;  	       // random(), most recently tx pkt 
  txw_trail_ = txw_lead_ + 1;  // random(), oldest avail. pkt 
  spm_sqn_ = 0;
}

int
PGMSndAgent::command(int argc, const char*const* argv)
{
  if (argc == 2)
    if ( !strcmp(argv[1], "start") ) {
      start();
      return TCL_OK;
    }
  return PGMAgent::command(argc, argv);
}

void 
PGMSndAgent::start(void) 
{
  double now = Scheduler::instance().clock();
  last_update_num_bytes_ = now ;
  odata_time_of_last_insert_ = now - txw_secs_  ;
  numbytes_ = txw_max_rte_ ;
  timeout(AMBIENT_SPM) ;
}

void 
PGMSndAgent::recv(Packet* p, Handler* )
{
  hdr_cmn *ch = hdr_cmn::access(p);
  hdr_pgm *ph = (hdr_pgm *) p->access(off_pgm_);
  u_int seq ;

  if ( ph->options() & PGM_OPT_PRESENT ) 
    option_handler(p) ;
  
  switch(ch->ptype()) {
  case PT_CBR:
  case PT_EXP:
  case PT_PARETO:
    odata_curr_.enque(p) ;   // prepare the xmit queue.
    data_move() ;            // send thru the traffic shaper. 
    return;

  case PT_PGM_NAK: 
    pkts_rcvd_++ ;
    nak_rcvd_++;
    seq = ph->sqn() ;
    /* multicast an NCF only for in-window requests */
    if ( pgm_seq_geq(seq, txw_trail_) && pgm_seq_leq(seq, txw_lead_) ) {
       hdr_pgm_nak * nakh = (hdr_pgm_nak *) p->access(off_pgm_nak_) ;

       // allocate a new NCF packet 
       size_ = HDR_SIZE_PGM_NCF ;
       type_ = PT_PGM_NCF ;
       Packet * ncf = allocpkt() ;
       hdr_pgm * ph2 = (hdr_pgm *) ncf->access(off_pgm_) ;
       hdr_pgm_ncf * ncfh = (hdr_pgm_ncf *) ncf->access(off_pgm_ncf_) ;
       ncfh->src_nla() = nakh->src_nla() ;
       ph2->sqn() = seq ;
       print_xplot(p) ;
       ph2->options() = 0 ;
       ph2->opttype() = 0 ;
       buildoptions(ncf, PT_PGM_NCF) ; // add options

       // send the NCF
       output(ncf, PT_PGM_NCF) ; 

       // schedule the RDATA
       Packet * r = odata_q_.lookup(seq - txw_trail_) ; 
       rdata_q_.insert(r->copy()) ;	 

       // try to send 
       data_move() ;		 
    } else {
#ifdef PGMCC
       fprintf(stderr, "-- [%5.3f] NAK outside window %u<=%u<=%d\n", 
              Scheduler::instance().clock(), txw_trail_, seq, txw_lead_) ;
#endif PGMCC
	}
    Packet::free(p) ;
    return; 

  case PT_PGM_NNAK: 
    pkts_rcvd_++ ;
    nnak_rcvd_++ ;
    // yet unimplemented! 
    Packet::free(p) ;
    return; 

  case PT_PGM_ODATA:
  case PT_PGM_RDATA:
  case PT_PGM_SPM:
    fprintf(stderr, "-- [%5.3f] PGMSrc: Src received its own packet\n",
            Scheduler::instance().clock() ) ;
    Packet::free(p) ;
    return ;

  case PT_PGM_NCF:
    fprintf(stderr, "-- [%5.3f] PGMSrc: Src received an NCF packet\n", 
            Scheduler::instance().clock() ) ;
    Packet::free(p) ;
    return ;

  default:
    fprintf(stderr, "-- [%5.3f] PGMSrc: Unknown packet type\n", 
            Scheduler::instance().clock() ) ;
    Packet::free(p) ;
  }
}

void
PGMSndAgent::next_packet() 
{
	Packet * p ;
	hdr_cmn * ch ;

	if ( txw_max_rte_ ) {
		p = ( rdata_q_.length() )? rdata_q_.head() : odata_curr_.head() ;
		if ( !p ) {
			next_packet_timer_.force_cancel() ;
			return ;
		}
		ch = hdr_cmn::access(p);
		double t = (ch->size_ - numbytes_)/txw_max_rte_ ;
		next_packet_timer_.resched(t) ;
	}
}

int
PGMSndAgent::can_tx()
{
	return (odata_curr_.length() > 0 && numbytes_ >= 0) ;
} 


/*
 * data_move() 
 * traffic shaper, sends odata/rdata packets.
 * bandwidth budget includes odata and rdata only.
 */
void
PGMSndAgent::data_move() 
{
  Packet * p ;
  hdr_cmn * ch ;
  hdr_pgm * ph ;

  // free expired rdata
  while ( rdata_q_.head() != NULL ) {
    p = rdata_q_.head() ;
    ph = (hdr_pgm *) p->access(off_pgm_) ;
    if ( pgm_seq_geq(ph->sqn(),txw_trail_) ) 
       break ;
    Packet::free(rdata_q_.deque()) ;
  }
 
  // update numbytes_
  if ( txw_max_rte_ && numbytes_ < 0 ) {
    double now = Scheduler::instance().clock() ;
    numbytes_ += (now - last_update_num_bytes_)*txw_max_rte_ ;
    last_update_num_bytes_ = now ;
  }

  // transmit some rdata
  while ( rdata_q_.length() != 0 && numbytes_ >= 0 ) {
     p = rdata_q_.deque() ;
     ch = hdr_cmn::access(p);
     buildoptions(p, PT_PGM_RDATA) ;
     if ( txw_max_rte_ ) 
        numbytes_ -= ch->size() ;
     output(p, PT_PGM_RDATA) ;
  }

  // transmit some odata 
  while ( can_tx() ) {
     p = odata_curr_.deque() ;
     ch = hdr_cmn::access(p);
     ph = (hdr_pgm *) p->access(off_pgm_) ;
     ch->size() += HDR_SIZE_PGM_ODATA ;
     ph->options() = 0 ;
     ph->opttype() = 0 ; 
     buildoptions(p, PT_PGM_ODATA) ;
     if ( txw_max_rte_ ) 
        numbytes_ -= ch->size() ;
     output(p, PT_PGM_ODATA) ;   
     if ( infinite_send_ ) {
        // infinite data buffer, so we enqueue 
        // a new packet in odata_curr_
        size_ = payload_ ;
        type_ = PT_PGM_ODATA ;
        odata_curr_.enque(allocpkt()) ;
     } else if ( app_waiting_ ) {
        // if the application is blocked on the sendmsg,
        // queue its packet in odata_curr_ and wake it up.
        odata_curr_.enque(pkt_waiting_) ;
        app_->resume() ;
    }
  }

  // schedule next packet 
  next_packet() ;
}

void
PGMSndAgent::output(Packet* p, packet_t type)
{

        hdr_cmn *ch = hdr_cmn::access(p);
 	hdr_ip *ih = hdr_ip::access(p);
  	hdr_pgm *ph = (hdr_pgm *) p->access(off_pgm_);
    hdr_pgm_odata *poh ;
    hdr_pgm_rdata *prh ;
  	hdr_pgm_spm *psh ;

  	ch->ptype() = type ;
  	ch->timestamp() = Scheduler::instance().clock() ;
  	// ip header
  	ih->dst() = dst_ ;
  	ih->src() = here_;
  	// pgm header
  	ph->tsi() = here_.addr_ ;

  	switch (type) {
  	case PT_PGM_ODATA: 
    	poh = (hdr_pgm_odata *) p->access(off_pgm_odata_);
    	ph->sqn() = ++txw_lead_ ;
    	poh->trail() = txw_trail_ ;
    	odata_sent_++;
#ifdef PGMCC
    	fprintf(stderr, "-- [%5.3f] new ODATA: seq %d trail %d\n",
            	Scheduler::instance().clock(), ph->sqn(), poh->trail()) ;
#endif
    	/* store packet in the window queue for retransmission */
    	if ( trail_advance_policy_ == TRAIL_ADVANCE_TIMER ) {
       		double now = Scheduler::instance().clock() ;
       		double lifetime = now - odata_time_of_last_insert_ ;
       		if ( lifetime > txw_secs_ )
          		lifetime = txw_secs_ ;
       		if ( odata_q_.empty() )
          		txw_advance_timer_.resched(lifetime) ;
       		Packet * r = p->copy() ;
       		hdr_pgm * rh = (hdr_pgm *) r->access(off_pgm_) ;
       		rh->options() = 0 ;
       		rh->opttype() = 0 ;
       		odata_q_.enque(r, lifetime) ;   
       		odata_time_of_last_insert_ = now ;
    	} // TRAIL_ADVANCE_DATA is not yet implemented
    	break;

 	case PT_PGM_RDATA: 
    	prh = (hdr_pgm_rdata *) p->access(off_pgm_rdata_);
    	prh->trail() = txw_trail_ ;
    	rdata_sent_++;
#ifdef PGMCC
    	fprintf(stderr, "-- [%5.3f] new RDATA: seq %d trail %d\n",
            	Scheduler::instance().clock(), ph->sqn(), prh->trail()) ;
#endif
    	break;

  	case PT_PGM_SPM:   
  		psh = (hdr_pgm_spm *) p->access(off_pgm_spm_);
    	ph->sqn() = spm_sqn_++;
    	psh->trail() = txw_trail_;
    	psh->lead() = txw_lead_;
    	psh->path_nla() = here_.addr_;
    	spm_sent_++;
#ifdef PGMCC	
    	fprintf(stderr, "-- [%5.3f] new SPM: seq %d trail %d lead %d\n",
            	Scheduler::instance().clock(), ph->sqn(),
		        psh->trail(), psh->lead()) ;
#endif
    	break;
  	case PT_PGM_NCF: 
    	ih->ttl() = 1 ;
    	ncf_sent_++; 
#ifdef PGMCC
    	fprintf(stderr, "-- [%5.3f] new NCF: seq %d \n",
            	Scheduler::instance().clock(), ph->sqn()) ;
#endif
    	break;

  	default: // Severe error
    	fprintf(stderr, "[%5.3f] PGM Error: Cannot process packet of type %d\n", 
            Scheduler::instance().clock(), ch->ptype()) ;
    	exit(EXIT_FAILURE) ;
  	} 

  	pkts_sent_++;
  	print_xplot(p) ;
  	send(p, 0);
}

void
PGMSndAgent::timeout(int tno)
{
  switch(tno) { 
  case NEXT_PACKET:
    data_move() ;
    break ;
  case AMBIENT_SPM: {
    // allocate new SPM packet 
    size_ = HDR_SIZE_PGM_SPM ;
    type_ = PT_PGM_SPM ;
    Packet *p = allocpkt() ;
    hdr_pgm * ph = (hdr_pgm *) p->access(off_pgm_) ;
    ph->options() = 0 ;
    ph->opttype() = 0 ;
    buildoptions(p, PT_PGM_SPM) ;
    output(p, PT_PGM_SPM);
    spm_ambient_timer_.resched(spm_ambient_time_);
    }
    break ; 
  case TXW_ADVANCE:
    txw_trail_++ ;
    Packet::free(odata_q_.deque()) ;
    if ( !odata_q_.empty() ) 
       txw_advance_timer_.resched(odata_q_.time()) ;
    break ;
  default: // severe error
    fprintf(stderr, "Error: Unknown timer expired\n") ;
    exit(EXIT_FAILURE) ;
  }
}   

void
PGMSndAgent::buildoptions(Packet * p, packet_t type) 
{
  	hdr_cmn * ch = hdr_cmn::access(p);
  	hdr_pgm * ph = (hdr_pgm *) p->access(off_pgm_) ;

  	if ( optjoin_ && type != PT_PGM_NCF ) {
     	ph->options() |= PGM_OPT_PRESENT ;
  	   	ch->size() += OPT_JOIN_SIZE ; 
     	ph->opttype() |= PGM_OPT_JOIN ;
     	ph->opt_join_min() = txw_trail_ ;
  	}

  	// if options are present, put total length field
  	if ( ph->options() & PGM_OPT_PRESENT ) 
     	ch->size() += OPT_LENGTH_SIZE ;
}

void
PGMSndAgent::option_handler(Packet * ) {
  // currently, no options are supported.
}

void
PGMSndAgent::sendmsg(int nbytes, const char *) 
{
  if (nbytes == -1) {
#ifdef PGMCC
     fprintf(stderr, "-- [%5.3f] Infinite send\n",
             Scheduler::instance().clock()) ;
#endif
     infinite_send_ = TRUE ;
     // allocate first packet to start infinite loop
     size_ = payload_ ;
     type_ = PT_PGM_ODATA ;
     odata_curr_.enque(allocpkt()) ;
     data_move() ;
     return ;
  }

  // allocate new data packet 
  size_ = nbytes ;
  type_ = PT_PGM_ODATA ;
  Packet * p = allocpkt() ;

  if ( odata_curr_.length() >= sendspace_ ) {
     // not enough sendspace_ for new packet,  
     // block application on sendmsg,  
     // and store the new packet
     app_waiting_ = TRUE ; 
     pkt_waiting_ = p ;
     return ;
  }

  // prepare xmit queue
  odata_curr_.enque(p) ;
  // send thru the traffic shaper
  data_move() ;
  // signal application we are ready for new data
  if ( !app_waiting_ ) 
     app_->resume() ;
}
