/*
 * pgmsnd.h - 000330
 * Copyright (c) 2000 Gianluca Iannaccone, Luigi Rizzo
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR `AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#ifndef ns_pgmsnd_h
#define ns_pgmsnd_h

#include "agent.h"
#include "packet.h"
#include "queue.h"
#include "my_pgm.h"

class PGMWindowPacket {  
public:  
  PGMWindowPacket(Packet * p, double time) : pkt_(p), lifetime_(time) {
     pkt_->next_ = NULL ;
  }
  Packet * pkt() { return pkt_ ; } 
  double lifetime() { return lifetime_ ; } 
  PGMWindowPacket * next_ ;
protected:
  Packet * pkt_ ;
  double lifetime_ ;  
} ;

// PGMTransmitWindow class implements the transmit window 

class PGMTransmitWindow {
public:
  PGMTransmitWindow() : head_(NULL), tail_(NULL) {} 
  ~PGMTransmitWindow() {} 
  void enque(Packet * p, double time) { 
     PGMWindowPacket * n = new PGMWindowPacket(p, time) ;
     if ( !tail_ ) 
        head_= tail_= n;
     else {
        tail_->next_= n;
        tail_= n;
     }
    tail_->next_= NULL;
  }
  Packet* deque() {
     if ( !head_ ) 
        return NULL;
     PGMWindowPacket * n = head_;
     head_= n->next_; 
     if (n == tail_) 
        head_= tail_= NULL; 
     Packet * p = n->pkt() ;
     delete n ;
     return p ;
  }
  Packet* lookup(int n) {
     for (PGMWindowPacket * p = head_; p != NULL; p = p->next_) {
        if (--n < 0)
           return (p->pkt()) ;
     }
     return (NULL);
  }
  int empty() { return ( head_ == NULL ) ; }
  double time() { return head_->lifetime() ; } 
protected:
  PGMWindowPacket* head_ ;
  PGMWindowPacket* tail_ ;
} ;


class PGMOrderedQueue : public PacketQueue {
public:         
   virtual void insert(Packet* p) {
       
    if (len_ == 0) { // empty list 
      enque(p); 
      return;   
    }           

    Packet * q, * r ;
    hdr_pgm * ph = hdr_pgm::access(p) ;
    u_int seq = ph->sqn() ;
    for ( q = NULL, r = head_ ; r != NULL ; q = r, r = r->next_ ) {
       ph = hdr_pgm::access(r) ;
       if ( ph->sqn() == seq ) 
          return; /* nothing to do, entry already existing */
       else if ( ph->sqn() > seq )
          break ;
    } 

    if (q == NULL) { // insert p at the head of the list
      p->next_ = head_;
      head_ = p;
      len_++;   
      return;           
    }                           
                
    if (q->next_ == NULL) { // insert p at the tail
      enque(p);
      return;
    }   
        
    // insert p in the middle of the list
    p->next_ = q->next_;
    q->next_ = p;
    len_++;
  }     
};              


class PGMSndAgent : public PGMAgent {
public:
  	PGMSndAgent();
  	virtual void recv(Packet *, Handler *) ;
  	virtual void sendmsg(int, const char * flags = 0) ;
  	virtual int command(int, const char*const*) ;
  	virtual void timeout(int) ;
  	virtual void output(Packet *, packet_t) ;

protected:
  	virtual void start() ;	 			// called from OTcl to start
  	virtual void data_move() ; 			// traffic shaper
  	virtual int can_tx() ;				// enough bandwidth ? 
  	virtual void next_packet() ; 		// schedule packets tx
  	virtual void buildoptions(Packet *, packet_t) ; // fill pkts with options
  	virtual void option_handler(Packet *) ; // option handler

 	// sender state variables: max rate, trail seqno, lead seqno
  	double txw_max_rte_ ; 				// bytes/sec
  	u_int txw_trail_, txw_lead_ ;  

  	PacketQueue odata_curr_ ;			// tx queue, next pkts to send
  	PGMTransmitWindow odata_q_ ;		// tx window
  	PGMOrderedQueue rdata_q_ ;  		// RDATA queue (ordered by seqno)

  	int sendspace_ ;					// sender space in packets 

  	// next fields are used to implement the expire of odata
  	// in timer-advance mode.
  	int trail_advance_policy_ ;
#define TRAIL_ADVANCE_TIMER 	1		// advance with time
#define TRAIL_ADVANCE_DATA	2			// advance with data 
#define TRAIL_ADVANCE_USER	3			// advance controlled by apps.
  	double txw_secs_ ;					// interval for which repair is avail. 
  	double odata_time_of_last_insert_ ;   

  	u_int spm_sqn_;
  	double spm_ambient_time_ ; 			// Ambient SPM 
										// Heartbeat SPM not implemented

  	// for dealing with the application 
  	int app_waiting_ ;					// app is blocked on sendmsg
  	Packet * pkt_waiting_ ;				// pkt that app want to send 
  	int infinite_send_ ; 				// infinite tx mode
  	int payload_ ;						// pkt size for infinite tx mode  

  	double numbytes_ ;					// amount of data i can send next time
  	double last_update_num_bytes_ ;  	// time since last update of numbytes_

  	// options 
  	int optjoin_ ;

  	// timers
  	PGMTimer next_packet_timer_;		// send next packet
  	PGMTimer txw_advance_timer_;		// advance transmit window
  	PGMTimer spm_ambient_timer_;		// send an ambient spm
};

#endif
